"""
ARNICA
======

Provides a python wrapper to setup, run and post-process AVBP simulations.

Available subpackages
---------------------
utils
    A collection of tools for geometrical problems
phys
    A collection of tools for physical problems, essentially around CHT configurations
solvers_2d
    Small 2d structured FD solvers in(Dperecated)
"""

from arnica import utils
from arnica import phys
from arnica import solvers_2d
